/*
 * Copyright (c) 1990 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <stdio.h>
#include "dixie.h"

#include <quipu/attrvalue.h>
#include <quipu/ds_error.h>
#include <quipu/commonarg.h>
#include <quipu/list.h>
#include <quipu/dua.h>

#define NULL	0

#define LIST_END(x)	{ dn_free(thedn); \
			return(x); }

extern int	debug;
extern int	bound;
extern int	stream_connected;
extern short	timelimit, sizelimit;

/* args for the read call */
struct ds_list_arg	list_arg = { default_common_args };
struct ds_list_result	list_result;
struct DSError		list_error;

listop(bp, rp, rpsize, dontderef)
char	*bp, *rp;
int	rpsize;
char	dontderef;
{
	PS		ps;
	DN		thedn = NULLDN;
	struct subordinate *sub;
	int		i;
	short		entries;
	int		timeproblem, sizeproblem;

	list_arg.lsa_common.ca_servicecontrol.svc_timelimit = timelimit;
	list_arg.lsa_common.ca_servicecontrol.svc_sizelimit = sizelimit;
	if ( dontderef )
		list_arg.lsa_common.ca_servicecontrol.svc_options |=
		    SVC_OPT_DONTDEREFERENCEALIAS;
	else
		list_arg.lsa_common.ca_servicecontrol.svc_options &=
		    ~SVC_OPT_DONTDEREFERENCEALIAS;
	rp[0] = rp[1] = rp[2] = '\0';
	if ( debug & 2 ) printf("List (%s)\n", bp);

	/* 
	 * First convert the dn given into a DN structure
	 */

	if ( *bp == '@') bp++;
	if ( (thedn = str2dn(bp)) == NULLDN ) {
		if ( debug > 1 ) printf("listop: str2dn (%s) failed\n",bp);
		if ( *bp != '\0' )
			LIST_END(DXI_BADDN);
	}
	list_arg.lsa_object = thedn;

	/* do the actual list */
	timeproblem = sizeproblem = 0;
	for ( i = 0; i < 2; i++ ) {
		if ( ds_list(&list_arg, &list_error, &list_result) == DS_OK )
			break;
		if ( debug & 2 ) printf("ds_list failed\n");

		if ( list_error.dse_type == DSE_REFERRAL )
			if ( chase_referral( &list_error ) != OK )
				break;
			else
				continue;

		if ( list_error.dse_type != DSE_REMOTEERROR
		    || stream_connected )
			break;

		(void) nullbind();

		if ( bound == 0 )
			break;
	}

	if ( debug & 2 )
		printf("list successful - checking for errors\n");

	switch ( list_error.dse_type ) {
	case DSE_REMOTEERROR:
		if ( debug & 2 ) {
			printf("\tservice error\n");
			error( &list_error );
		}
		LIST_END(DXI_DSA_DOWN);
	case DSE_REFERRAL:
		if ( debug & 2 ) {
			printf("\treferral error\n");
			error( &list_error );
		}
		LIST_END(DXI_REFER_ERROR);
	case DSE_NAMEERROR:
		if ( debug & 2 ) {
			printf("\tname error\n");
			error( &list_error );
		}
		LIST_END(DXI_NAME_ERROR);
	case DSE_SERVICEERROR:
		if ( debug & 2 ) {
			printf("\tservice error\n");
			error( &list_error );
		}
		if ( list_error.dse_un.dse_un_service.DSE_sv_problem
		    == DSE_SV_TIMELIMITEXCEEDED )
			timeproblem = 1;
		else if ( list_error.dse_un.dse_un_service.DSE_sv_problem
		    == DSE_SV_ADMINLIMITEXCEEDED)
			sizeproblem = 1;
		else
			LIST_END(DXI_ERROR)
		break;
	case DSE_NOERROR:
		if ( debug & 2 ) printf("\tno error\n");
		break;
	default:
		if ( debug & 2 ) {
			printf("\tsome unknown error (%d)\n");
			error( &list_error );
		}
		LIST_END(DXI_ERROR);
	}

	timeproblem |= (list_result.lsr_limitproblem == LSR_TIMELIMITEXCEEDED);
	sizeproblem |= (list_result.lsr_limitproblem == LSR_SIZELIMITEXCEEDED
	    || list_result.lsr_limitproblem == LSR_ADMINSIZEEXCEEDED);

	if ((ps = ps_alloc(str_open)) == NULLPS) {
		if (debug) printf("ps_alloc failed\n");
		clean_exit(1);
	}
	if (str_setup(ps, rp + 2, rpsize - 2, 1) == NOTOK) {
		if (debug) printf("str_setup failed\n");
		clean_exit(1);
	}

	entries = 0;
	for ( sub = list_result.lsr_subordinates; sub != NULLSUBORD;
	    sub = sub->sub_next ) {
		if ( debug & 2 ) {
			printf("subordinate (%s)\n",
			    sub->sub_rdn->rdn_av.av_struct);
			fflush(stdout);
		}

		ps_print(ps, "\003");
		rdn_print(ps, sub->sub_rdn, EDBOUT);

		entries++;
	}
	*ps->ps_ptr = '\0';
	ps_free(ps);

	(void) bcopy((char *) &entries, rp, sizeof(short));

	LIST_END(timeproblem ? DXI_TIMELIMIT :
	    sizeproblem ? DXI_SIZELIMIT : DXI_SUCCESS);
}
